package net.sergeych.kiloparsec.adapter

import kotlinx.coroutines.*
import kotlinx.coroutines.channels.Channel
import kotlinx.coroutines.channels.ClosedReceiveChannelException
import net.sergeych.bintools.toDump
import net.sergeych.crypto.encodeVarUnsigned
import net.sergeych.crypto.readVarUnsigned
import net.sergeych.kiloparsec.Transport
import net.sergeych.mp_logger.LogTag
import net.sergeych.mp_logger.warning
import net.sergeych.mp_tools.globalLaunch
import java.net.InetSocketAddress
import java.net.StandardSocketOptions.TCP_NODELAY
import java.nio.ByteBuffer
import java.nio.channels.AsynchronousSocketChannel
import kotlin.coroutines.cancellation.CancellationException
import kotlin.coroutines.suspendCoroutine

private val log = LogTag("ASTD")

/**
 * Prepend block with its size, varint-encoded
 */
private fun encode(block: UByteArray): ByteArray
    = (encodeVarUnsigned(block.size.toUInt()) + block).toByteArray()

/**
 * Convert asynchronous socket to a [Transport.Device] using non-blocking nio,
 * in a coroutine-effective manner. Note that it runs coroutines to read/write
 * to the socket in a global scope.These are closed when transport is closed
 *  or the socket is closed, for example, by network failure.
 */
suspend fun asyncSocketToDevice(socket: AsynchronousSocketChannel): InetTransportDevice {
    val deferredDevice = CompletableDeferred<InetTransportDevice>()
    globalLaunch {
        coroutineScope {
            fun stop() {
                cancel()
            }
            socket.setOption(TCP_NODELAY, true)
            // socket input is to be parsed for blocks, so we receive bytes
            // and decode them to blocks
            val input = Channel<UByte>(1024)
            // copy from socket to input:
            launch {
                val data = ByteArray(1024)
                val inb = ByteBuffer.wrap(data)
                while (isActive) {
                    inb.position(0)
                    val size: Int = suspendCoroutine { continuation ->
                        socket.read(inb, continuation, IntCompletionHandler)
                    }
                    if (size < 0) stop()
                    else {
                        println("recvd:\n${data.sliceArray(0..<size).toDump()}\n------------------")
                        for (i in 0..<size) input.send(data[i].toUByte())
                    }
                }
            }
            // output is blocks, so we sent blocks:
            val outputBlocks = Channel<UByteArray>()
            // copy from output to socket:
            launch {
                try {
                    while (isActive) {
                        // wait for the first block to send
                        var data = encode(outputBlocks.receive())
                        // if there are more, take them all (NO_DELAY optimization)
                        while (!outputBlocks.isEmpty)
                            data += encode(outputBlocks.receive())
                        // now send the aggregate:
                        val outBuff = ByteBuffer.wrap(data)
                        val cnt = suspendCoroutine { continuation ->
                            socket.write(outBuff, continuation, IntCompletionHandler)
                        }
                        // be sure it was all sent
                        if( outBuff.position() != data.size || cnt != data.size) {
                            throw RuntimeException("PArtial write!")
                        }
                    }
                } catch (_: ClosedReceiveChannelException) {
                    stop()
                }
            }
            // transport device copes with blocks:
            val inputBlocks = Channel<UByteArray?>()
            // decode blocks from a byte channel read from the socket:
            launch {
                try {
                    while (isActive) {
                        val size = readVarUnsigned(input)
                        if (size == 0u) log.warning { "zero size block is ignored!" }
                        else {
                            val block = UByteArray(size.toInt())
                            for (i in 0..<size.toInt()) {
                                block[i] = input.receive()
                            }
                            inputBlocks.send(block)
                        }
                    }
                } catch (_: CancellationException) {
                    inputBlocks.send(null)
                } catch (_: ClosedReceiveChannelException) {
                    inputBlocks.send(null)
                    stop()
                }
            }
//            SocketAddress.
            val addr = socket.remoteAddress as InetSocketAddress
            deferredDevice.complete(
                InetTransportDevice(inputBlocks, outputBlocks, JvmNetworkAddress(addr.address,addr.port)) { stop() }
            )
        }
        globalLaunch { socket.close() }
    }
    return deferredDevice.await()
}
